# -*- coding: utf-8 -*-
"""String Converter.

Module provides function for converting values to strings.

"""

# Standard Library Imports
import datetime
import decimal
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_string"]


def to_string(
    __value: object, /, default: Optional[str] = None
) -> Optional[str]:
    """Convert value to string.

    Args:
        value: Value to convert.

    Returns:
        String.

    """
    converter = StringConverter(default=default)
    result = converter(__value)
    return result


class StringConverter(BaseConverter):
    """Class implements a string converter.

    Args:
        default (optional): Default value. Default ``None``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: Optional[str] = None,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if default and not isinstance(default, str):  # type: ignore
            message = f"expected type 'str', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()

    @property
    def default(self) -> Any:  # pragma: no cover
        """Default value."""
        return self._default

    @default.setter
    def default(self, value: Any) -> None:  # pragma: no cover
        if not isinstance(value, str):  # type: ignore
            message = f"expected type 'str', got {type(value)} instead"
            raise TypeError(message)

        self._default = value


def str_from_bool(__value: bool, _: Optional[str] = None, /) -> str:
    """Convert boolean value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'bool'.

    """
    if not isinstance(__value, bool):  # type: ignore  # pragma: no cover
        message = f"expected type 'bool', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_date(__value: datetime.date, _: Optional[str] = None, /) -> str:
    """Convert date value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore  # pragma: no cover
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_datetime(
    __value: datetime.datetime, _: Optional[str] = None, /
) -> str:
    """Convert datetime value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore  # pragma: no cover
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_decimal(
    __value: decimal.Decimal, _: Optional[str] = None, /
) -> str:
    """Convert decimal value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore  # pragma: no cover
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_float(__value: float, _: Optional[str] = None, /) -> str:
    """Convert float value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):  # type: ignore  # pragma: no cover
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_int(__value: int, _: Optional[str] = None, /) -> str:
    """Convert integer value to ``str``.

    Args:
        __value: Value to convert to ``str``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore  # pragma: no cover
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = str(__value)
    return result


def str_from_str(
    __value: str, default: Optional[str] = None, /
) -> Optional[str]:
    """Convert string value to ``str``.

    Args:
        __value: Value to convert to ``str``.
        default (optional): Default value. Default ``None``.

    Returns:
        String.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to ``str``.

    """
    if not isinstance(__value, str):  # type: ignore  # pragma: no cover
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    result = str(__value)
    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., Optional[str]]] = {
    bool: str_from_bool,
    datetime.date: str_from_date,
    datetime.datetime: str_from_datetime,
    decimal.Decimal: str_from_decimal,
    float: str_from_float,
    int: str_from_int,
    str: str_from_str,
}



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib
from types import ModuleType
from typing import Literal
from typing import Optional
from typing import overload

__all__ = [
    "import_module",
    "raise_for_instance",
]


@overload
def import_module(
    name: str,
    package: Optional[str] = None,
    *,
    required: Literal[True] = True,
) -> ModuleType: ...


@overload
def import_module(
    name: str,
    package: Optional[str] = None,
    *,
    required: Literal[False] = False,
) -> Optional[ModuleType]: ...


def import_module(
    name: str,
    package: Optional[str] = None,
    *,
    required: bool = True,
) -> Optional[ModuleType]:
    """Import module.

    Args:
        name: Name of module.
        package (optional): Package. Default ``None``.
        required: Whether package is required. Default ``True``.

    Returns:
        Module.

    """
    try:
        result = importlib.import_module(name, package)

    except (ImportError, ModuleNotFoundError) as error:
        if required is True:
            raise error

        return None

    return result


def raise_for_instance(__obj: object, __class: type) -> None:
    """Raise error when object is not instance of provided class.

    Args:
        __obj: Object for which to check class.
        __class: Class for which to check.

    Raises:
        TypeError: when object is not an instance of class.

    """
    if not isinstance(__obj, __class):
        expected = f"expected type '{__class.__name__!s}'"
        actual = f"got {type(__obj)} instead"
        message = ", ".join([expected, actual])
        raise TypeError(message)



# -*- coding: utf-8 -*-

# Standard Library Imports
import re

__all__ = ["parse_number"]


def parse_number(__value: str, /) -> float:
    """Parse number from string.

    Args:
        __value: String that contains a number.

    Returns:
        String representation of number.

    Raises:
        TypeError: when value is not type 'str'.

    """
    if not isinstance(__value, str):  # type: ignore  # pragma: no cover
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    representation = re.sub(r"[^0-9a-zA-Z.]+", r"", value)
    number = re.sub(r"(\d+)\s?[a-z]*", r"\1", representation, flags=re.I)
    result = float(number)
    return result



# -*- coding: utf-8 -*-
"""CSV File Wrappers."""

# Standard Library Imports
from __future__ import annotations
import csv
import os
from typing import Any
from typing import Collection
from typing import Dict
from typing import IO
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

# Local Imports
from .abstract_file_wrappers import AbstractDirectoryWrapper
from .abstract_file_wrappers import AbstractFileWrapper
from .abstract_file_wrappers import AbstractTextWrapper
from .abstract_file_wrappers import AbstractIOWrapper
from ..utils import converters
from .. import helpers
from .. import settings
from .. import utils

__all__ = [
    "AbstractCsvWrapper",
    "CsvDirectoryWrapper",
    "CsvFileWrapper",
    "CsvIOWrapper",
]


class AbstractCsvWrapper(AbstractTextWrapper):
    """Represents an abstract wrapper class for `.csv` files."""

    @property
    def delimiter(self) -> str:
        """Delimiter."""
        return self._delimiter

    @delimiter.setter
    def delimiter(self, value: Any) -> None:
        helpers.raise_for_instance(value, str)
        self._delimiter: str = value

    @property
    def dialect(self) -> Union[csv.Dialect, str]:
        """Dialect."""
        return self._dialect

    @dialect.setter
    def dialect(self, value: Any) -> None:
        if not isinstance(value, (csv.Dialect, str)):
            expected = "expected type 'Dialect' or 'str'"
            actual = f"got {type(value)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        self._dialect: Union[csv.Dialect, str] = value

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, value: Any) -> None:
        helpers.raise_for_instance(value, Sequence)
        self._fieldnames: Sequence[str] = value

    @property
    def quotechar(self) -> str:
        """Quote character."""
        return self._quotechar

    @quotechar.setter
    def quotechar(self, value: Any) -> None:
        helpers.raise_for_instance(value, str)
        self._quotechar: str = value

    def _init_csv_io_wrapper(self, __file: IO[Any], /) -> CsvIOWrapper:
        """Initialize I/O wrapper for `.csv` file.

        Args:
            __file: File-like object.

        Returns:
            I/O wrapper instance.

        """
        result = CsvIOWrapper(__file)
        setattr(result, "_context", self)
        return result


class CsvDirectoryWrapper(AbstractCsvWrapper, AbstractDirectoryWrapper):
    """Implements a wrapper for `.csv` files in a directory.

    Args:
        directory: Directory from which to load `.csv` file(s).
        delimiter (optional): Delimiter. Default `,`.
        dialect (optional): Dialect. Default `excel`.
        encoding (optional): File encoding. Default `utf-8`.
        fieldnames (optional): Fieldnames. Default ``None``.
        newline (optional): Newline character. Default ``None``.
        quotechar (optional): Quote character. Default ``"``.
        read_only (optional): Whether file is read only. Default ``False``.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        delimiter: str = settings.DEFAULT_CSV_DELIMITER,
        dialect: Union[csv.Dialect, str] = settings.DEFAULT_CSV_DIALECT,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        fieldnames: Optional[Sequence[str]] = None,
        newline: str = settings.DEFAULT_CSV_NEWLINE,
        quotechar: str = settings.DEFAULT_CSV_QUOTECHAR,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            directory,
            encoding=encoding,
            extension=settings.CSV_EXTENSION,
            newline=newline,
            read_only=read_only,
        )
        self.delimiter = delimiter
        self.dialect = dialect
        self.fieldnames = fieldnames or []
        self.quotechar = quotechar

    def open(self, filename: str, /, mode: str = "r") -> CsvIOWrapper:
        """Open a `.csv` file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """
        file = super().open(filename, mode=converters.to_text_file_mode(mode))
        result = self._init_csv_io_wrapper(file)
        return result


class CsvFileWrapper(AbstractCsvWrapper, AbstractFileWrapper):
    """Implements a wrapper for `.csv` files.

    Args:
        filepath: Path to `.csv` file.
        delimiter (optional): Delimiter. Default `,`.
        dialect (optional): Dialect. Default `excel`.
        encoding (optional): File encoding. Default `utf-8`.
        fieldnames (optional): Fieldnames. Default ``None``.
        newline (optional): Newline character. Default ``None``.
        quotechar (optional): Quote character. Default ``"``.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        ValueError: when `filepath` is not a `.csv` file.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        delimiter: str = settings.DEFAULT_CSV_DELIMITER,
        dialect: Union[csv.Dialect, str] = settings.DEFAULT_CSV_DIALECT,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        fieldnames: Optional[Sequence[str]] = None,
        newline: str = settings.DEFAULT_CSV_NEWLINE,
        quotechar: str = settings.DEFAULT_CSV_QUOTECHAR,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            filepath,
            encoding=encoding,
            newline=newline,
            read_only=read_only,
        )
        utils.raise_for_extension(filepath, settings.CSV_EXTENSION)

        self.delimiter = delimiter
        self.dialect = dialect
        self.fieldnames = fieldnames or []
        self.quotechar = quotechar

    def open(self, mode: str = "r") -> CsvIOWrapper:
        """Open the `.csv` file and return a file object.

        Args:
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        file = super().open(converters.to_text_file_mode(mode))
        result = self._init_csv_io_wrapper(file)
        return result


class CsvIOWrapper(AbstractIOWrapper):
    """Implements a I/O wrapper for `.csv` files."""

    def __init__(self, __file: IO[Any]) -> None:
        self._file = __file
        self._context: Optional[AbstractCsvWrapper] = None

    @property
    def file(self) -> IO[Any]:
        """File."""
        return self._file

    @property
    def closed(self) -> bool:
        """Whether file is closed."""
        return self._file.closed

    @property
    def context(self) -> Union[AbstractCsvWrapper, CsvIOWrapper]:
        """Context."""
        return self._context or self

    @property
    def delimiter(self) -> Optional[str]:
        """Delimiter."""
        result = getattr(self.context, "_delimiter", None)
        return result

    @delimiter.setter
    def delimiter(self, value: Any) -> None:
        helpers.raise_for_instance(value, str)
        setattr(self.context, "_delimiter", value)

    @property
    def dialect(self) -> Optional[Union[csv.Dialect, str]]:
        """Dialect."""
        result = getattr(self.context, "_dialect", None)
        return result

    @dialect.setter
    def dialect(self, value: Any) -> None:
        if not isinstance(value, (csv.Dialect, str)):
            expected = "expected type 'Dialect' or 'str'"
            actual = f"got {type(value)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        setattr(self.context, "_dialect", value)

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        result = getattr(self.context, "_fieldnames", [])
        return result

    @fieldnames.setter
    def fieldnames(self, value: Any) -> None:
        helpers.raise_for_instance(value, Sequence)
        setattr(self.context, "_fieldnames", value)

    @property
    def quotechar(self) -> Optional[str]:
        """Delimiter."""
        result = getattr(self.context, "_quotechar", None)
        return result

    @quotechar.setter
    def quotechar(self, value: Any) -> None:
        helpers.raise_for_instance(value, str)
        setattr(self.context, "_quotechar", value)

    @property
    def read_only(self) -> bool:
        """Whether read only."""
        return getattr(self._context, "read_only")

    def __enter__(self) -> CsvIOWrapper:
        return self

    def __exit__(self, *_) -> None:
        self.close()
        return

    def close(self) -> None:
        """Close `.csv` file."""
        self._file.close()

    def read(self, n: int = -1, /) -> str:
        """Read content of `.csv` file.

        Returns:
            Content.

        """
        result = self._file.read(n)
        return result

    def readline(self, limit: int = -1, /) -> str:
        """Read line from `.csv` file.

        Returns:
            Line.

        """
        result = self._file.readline(limit)
        return result

    def readlines(self, hint: int = -1, /) -> List[Any]:
        """Read lines from `.csv` file.

        Returns:
            Lines.

        """
        result = self._file.readlines(hint)
        return result

    def read_record(self) -> Dict[str, Any]:
        """Read record from `.csv` file.

        Returns:
            Record.

        """
        reader = self._get_record_reader()
        result: Dict[str, Any] = reader.read_record()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return result

    def read_records(self) -> List[Dict[str, Any]]:
        """Read records from `.csv` file.

        Returns:
            Records.

        """
        reader = self._get_record_reader()
        results = reader.read_records()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return results

    def _get_record_reader(self) -> _CsvRecordReader:
        """Get record reader for `.csv` file..

        Returns:
            Reader.

        """
        if not hasattr(self, "_record_reader"):
            self._start_record_reader()

        result: _CsvRecordReader = getattr(self, "_record_reader")
        return result

    def _start_record_reader(self) -> None:
        """Start record reader for `.csv` file."""
        reader = _CsvRecordReader(self)
        setattr(self, "_record_reader", reader)

    def read_row(self) -> List[str]:
        """Read row from `.csv` file.

        Returns:
            Row.

        """
        reader = self._get_row_reader()
        result = reader.read_row()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return result

    def read_rows(self) -> List[List[str]]:
        """Read rows from `.csv` file.

        Returns:
            Rows.

        """
        reader = self._get_row_reader()
        results = reader.read_rows()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return results

    def _get_row_reader(self) -> _CsvRowReader:
        """Get row reaader for `.csv` file.

        Returns:
            Reader.

        """
        if not hasattr(self, "_row_reader"):
            self._start_row_reader()

        result: _CsvRowReader = getattr(self, "_row_reader")
        return result

    def _start_row_reader(self) -> None:
        """Start row rwader for `.csv` file."""
        reader = _CsvRowReader(self)
        setattr(self, "_row_reader", reader)

    def write(self, content: str, /) -> int:  # type: ignore
        """Write content to `.txt` file.

        Args:
            content: Content.

        """
        result = self._file.write(content)
        return result

    def writelines(self, lines: Iterable[str], /) -> None:  # type: ignore
        """Write lines to `.txt` file.

        Args:
            lines: Lines.

        """
        result = self._file.writelines(lines)
        return result

    def write_header(self) -> None:
        """Write header."""
        writer = self._get_record_writer()
        writer.write_header()

    def write_record(self, record: Dict[str, Any], /) -> None:
        """Write record to `.csv` file.

        Args:
            record: Record to write.

        """
        writer = self._get_record_writer()
        writer.write_record(record)

    def write_records(self, records: Iterable[Dict[str, Any]], /) -> None:
        """Write records to `.csv` file.

        Args:
            records: Records to write.

        """
        writer = self._get_record_writer()
        writer.write_records(records)

    def _get_record_writer(self) -> _CsvRecordWriter:
        """Get record writer for `.csv` file.

        Returns:
            Writer.

        """
        if not hasattr(self, "_record_writer"):
            self._start_record_writer()

        result: _CsvRecordWriter = getattr(self, "_record_writer")
        return result

    def _start_record_writer(self) -> None:
        """Start record writer for `.csv` file."""
        writer = _CsvRecordWriter(self)
        setattr(self, "_record_writer", writer)

    def write_row(self, row: Iterable[Any], /) -> None:
        """Write row to `.csv` file.

        Args:
            row: Row to write.

        """
        writer = self._get_row_writer()
        writer.write_row(row)

    def write_rows(self, rows: Iterable[Iterable[Any]], /) -> None:
        """Write rows to `.csv` file.

        Args:
            rows: Rows to write.

        """
        writer = self._get_row_writer()
        writer.write_rows(rows)

    def _get_row_writer(self) -> _CsvRowWriter:
        """Get row `.csv` writer.

        Returns:
            Writer.

        """
        if not hasattr(self, "_writer"):
            self._start_row_writer()

        result: _CsvRowWriter = getattr(self, "_writer")
        return result

    def _start_row_writer(self) -> None:
        """Start standard `.csv` writer."""
        writer = _CsvRowWriter(self)
        setattr(self, "_writer", writer)


class _CsvRecordReader(Iterator[Any]):
    """Implements a record reader for `.csv` files."""

    def __init__(self, __wrapper: "CsvIOWrapper", /) -> None:
        self._reader = csv.DictReader(
            __wrapper.file,
            fieldnames=__wrapper.fieldnames or None,
            dialect=__wrapper.dialect or settings.DEFAULT_CSV_DIALECT,
            delimiter=__wrapper.delimiter or settings.DEFAULT_CSV_DELIMITER,
            quotechar=__wrapper.quotechar,
        )

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._reader.fieldnames or []

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._reader.fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._reader.line_num

    def __next__(self) -> Dict[str, Any]:
        result = self.read_record()
        return result

    def read_record(self) -> Dict[str, Any]:
        """Read record from `.csv` file.

        Returns
            Record.

        """
        result = next(self._reader)
        return result

    def read_records(self) -> List[Dict[str, Any]]:
        """Read records from `.csv` file.

        Returns:
            Records.

        """
        results = list(self._reader)
        return results


class _CsvRecordWriter:
    """Implements a record writer for `.csv` files."""

    def __init__(self, __wrapper: "CsvIOWrapper") -> None:
        self._writer = csv.DictWriter(
            __wrapper.file,
            fieldnames=__wrapper.fieldnames,
            dialect=__wrapper.dialect or settings.DEFAULT_CSV_DIALECT,
            delimiter=__wrapper.delimiter or settings.DEFAULT_CSV_DELIMITER,
            quotechar=__wrapper.quotechar,
        )
        self._row_num = 0

    @property
    def fieldnames(self) -> Collection[str]:
        """Fieldnames."""
        return self._writer.fieldnames

    @fieldnames.setter
    def fieldnames(self, fieldnames: Collection[str]) -> None:
        self._writer.fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def write_header(self) -> None:
        """Write header."""
        self._writer.writeheader()
        self._row_num += 1

    def write_record(self, __record: Dict[str, Any], /) -> None:
        """Write record to `.csv` file.

        Args:
            __record: Record to write.

        """
        self._writer.writerow(__record)
        self._row_num += 1

    def write_records(self, __records: Iterable[Dict[str, Any]], /) -> None:
        """Write records to `.csv` file.

        Args:
            __records: Records to write.

        """
        self._writer.writerows(__records)
        self._row_num += len(list(__records))


class _CsvRowReader(Iterator[Any]):
    """Implements a row reader for `.csv` files."""

    def __init__(self, __wrapper: "CsvIOWrapper") -> None:
        self._reader = csv.reader(
            __wrapper.file,
            __wrapper.dialect or settings.DEFAULT_CSV_DIALECT,
            delimiter=__wrapper.delimiter or settings.DEFAULT_CSV_DELIMITER,
            quotechar=__wrapper.quotechar,
        )
        self._fieldnames = __wrapper.fieldnames

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames or []

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._reader.line_num

    def __next__(self) -> List[str]:
        result = self.read_row()
        return result

    def read_row(self) -> List[str]:
        """Read row from `.csv` file.

        Returns
            Row.

        """
        if self.row_num == 0 and self.fieldnames:
            result = self._read_first_row()
            return result

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        result = next(self._reader)
        return result

    def read_rows(self) -> List[List[str]]:
        """Read rows from `.csv` file.

        Returns:
            Rows.

        """
        if self.row_num == 0 and self.fieldnames:
            results = [self._read_first_row(), *list(self._reader)]
            return results

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        results = list(self._reader)
        return results

    def _read_header(self) -> None:
        """Read header.

        Raises:
            RuntimeError: when `fieldnames` is not empty.
            RuntimeError: when not on first line of file.

        """
        if self.fieldnames:  # sanity check
            message = "'fieldnames' is not empty"
            raise RuntimeError(message)

        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        self.fieldnames = next(self._reader)

    def _read_first_row(self) -> List[str]:
        """Read first row.

        Args:
            reader: Reader.

        Raises:
            RuntimeError: when `reader` is not on first line of file.

        """
        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        first_row = next(self._reader)
        if not self._is_header(first_row):
            return first_row

        result = next(self._reader)
        return result

    def _is_header(self, row: Sequence[Any]) -> bool:
        """Check whether row is header.

        Args:
            row: Row.

        Returns:
            Whether row is header.

        Raises:
            RuntimeError: when `fieldnames` is empty.

        """
        if not self.fieldnames:  # sanity check
            message = "'fieldnames' is empty"
            raise RuntimeError(message)

        result = not set(self.fieldnames).difference(row)
        return result


class _CsvRowWriter:
    """Implements a row writer for `.csv` files."""

    def __init__(self, __wrapper: "CsvIOWrapper") -> None:
        self._writer = csv.writer(
            __wrapper.file,
            __wrapper.dialect or settings.DEFAULT_CSV_DIALECT,
            delimiter=__wrapper.delimiter or settings.DEFAULT_CSV_DELIMITER,
            quotechar=__wrapper.quotechar,
        )
        self._fieldnames = __wrapper.fieldnames
        self._row_num = 0

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames or []

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def write_header(self) -> None:
        """Write header."""
        self.write_row(self.fieldnames)

    def write_row(self, __row: Iterable[Any], /) -> None:
        """Write row to `.csv` file.

        Args:
            __row: Row to write.

        """
        self._writer.writerow(__row)
        self._row_num += 1

    def write_rows(self, __rows: Iterable[Iterable[Any]], /) -> None:
        """Write rows to `.csv` file.

        Args:
            __rows: Rows to write.

        """
        self._writer.writerows(__rows)
        self._row_num += len(list(__rows))



# -*- coding: utf-8 -*-

# Standard Library Imports
import datetime
import decimal
import json
import pathlib
from typing import Any
from typing import Dict
from typing import Hashable
from typing import Optional

# Local Imports
from .helpers import import_module

# Optional Third-Party Imports
np = import_module("numpy", required=False)
pd = import_module("pandas", required=False)

__all__ = ["JSONEncoder"]


class JSONEncoder(json.JSONEncoder):
    """Class implements a custom JSON encoder."""

    def encode(self, o: Any, *args: Any, **kwargs: Any) -> str:
        """Return a JSON string representation of a Python data structure."""
        result = super().encode(replace_nan(o), *args, **kwargs)
        return result

    def default(self, o: object) -> Any:
        if isinstance(o, datetime.datetime):
            return str(o)

        if isinstance(o, datetime.date):
            return o.strftime("%Y-%m-%d")

        if isinstance(o, decimal.Decimal):
            return float(o)

        if isinstance(o, pathlib.Path):
            return str(o.resolve())

        # Only check when numpy is installed.
        if np is not None and isinstance(o, getattr(np, "bool_")):
            return bool(o)

        # Only check when pandas is installed.
        if pd is not None and getattr(pd, "isnull")(o):
            return None

        return super().default(o)


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def replace_nan(__value: Any, repl: Optional[str] = None) -> Any:
    """Replace 'np.nan'.

    Args:
        __value: Value in which to replace 'np.nan'.
        repl: Value with which to replace 'np.nan'.

    Returns:
        Value.

    """
    if isinstance(__value, dict):
        result: Dict[Hashable, Any] = {
            k: replace_nan(v) for k, v in __value.items()  # type: ignore
        }
        return result

    if isinstance(__value, list):
        result = [replace_nan(i) for i in __value]  # type: ignore
        return result

    if pd and getattr(pd, "isnull")(__value):
        return repl

    return __value



# -*- coding: utf-8 -*-
"""Event class.

This module defines event classes. Events are simple data structures that work as
messages in the system, transmitting instructions from one part of the system to another.
Events distribute information whenever a specific event occurs so that subsequent actions
can be performed by processes in the system which are dependent on that information.
Events are always named using past-tense verb phrases.

"""

# Local Imports
from .message import AbstractMessage

__all__ = ["AbstractEvent"]


class AbstractEvent(AbstractMessage):
    """Class represents an abstract event."""



# -*- coding: utf-8 -*-

# Standard Imports
import abc

__all__ = ["AbstractListener"]


class AbstractListener(abc.ABC):
    """Represents an abstract listener."""

    @abc.abstractmethod
    def start(self) -> None:
        """Start listener."""
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self) -> None:
        """Stop listener."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Schema."""

# Standard Library Imports
from __future__ import annotations
import collections
from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Optional

__all__ = ["MapperSchema"]


# Constants
INWARD = "inward"
OUTWARD = "outward"


class MapperSchema(collections.UserDict):  # type: ignore
    """Implements a mapper schema.

    Args:
        __mappers: Mappers.

    """

    def __init__(self, __mappers: Dict[Hashable, Any]) -> None:
        if not isinstance(__mappers, dict):  # type: ignore
            message = f"expected type 'dict', got {type(__mappers)} instead"
            raise TypeError(message)

        self._data = standardize_attribute_mappers(__mappers)

    @property
    def data(self) -> Dict[Hashable, Any]:  # type: ignore
        """Schema data."""
        return self._data

    def get_converter(
        self,
        ref: Hashable,
        direction: str = OUTWARD,
        *,
        default: Callable[..., Any] = lambda x: x,  # type: ignore
    ) -> Optional[Callable[..., Any]]:
        """Get converter.

        Args:
            ref: Reference for attribute.
            direction (optional): Direction of converter. Default ``outward``.
            default (optional): Default converter. Default `lambda x: x`.

        Returns:
            Converter.

        """
        if direction not in (INWARD, OUTWARD):
            message = f"invalid direction for mapper: {direction}"
            raise ValueError(message)

        try:
            mapper = self.get_attribute_mapper(ref)
            result = mapper["converters"][direction] if mapper else None

        except KeyError:
            result = (
                mapper.get("converter") or default  # type: ignore
                if direction == OUTWARD
                else default
            )

        return result

    def get_attribute_mapper(
        self, ref: Hashable
    ) -> Optional[Dict[Hashable, Any]]:
        """Get attribute mapper.

        Args:
            ref: Reference for attribute.

        Returns:
            Attribute mapper.

        """
        try:
            result = self.data[ref]
        except KeyError:
            return None

        return result

    def get_attribute_names(self) -> Dict[Hashable, Hashable]:
        """Get attribute names.

        Returns:
            Attribute names.

        """
        results: Dict[Hashable, Hashable] = {
            get_attribute_name(attr): key
            for key, attr in self.data.items()
            if get_attribute_name(attr) is not None
        }  # type: ignore
        return results


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def standardize_attribute_mappers(
    __mappers: Dict[Hashable, Any], /
) -> Dict[Hashable, Any]:
    """Standardize attribute mappers.

    Args:
        __mappers: Attribute mappers.

    Returns:
        Attribute mappers.

    """
    results = {
        key: standardize_attribute_mapper(value)
        for key, value in __mappers.items()
    }
    return results


def standardize_attribute_mapper(__mapper: Any, /) -> Dict[Hashable, Any]:
    """Standardize attribute mapper.

    Args:
        __mapper: Attribute mapper.

    Returns:
        Attribute mapper.

    """
    if isinstance(__mapper, dict):
        return __mapper  # type: ignore

    if isinstance(__mapper, str):
        return {"map_to": __mapper}

    message = "invalid attribute mapper"
    raise ValueError(message)


# ----------------------------------------------------------------------------
# Selectors
# ----------------------------------------------------------------------------
def get_attribute_name(__mapper: Any, /) -> Optional[Hashable]:
    """Get mapped attribute.

    Args:
        __mapper: Attribute mapper.

    Returns:
        Mapped attribute.

    """
    if isinstance(__mapper, dict):
        return __mapper.get("map_to")  # type: ignore

    if isinstance(__mapper, str):
        return __mapper

    return None



# -*- coding: utf-8 -*-

# Standard Library Imports
from __future__ import annotations
from typing import Any

# Third-Party Imports
import redis

# Local Imports
from .abstract_publisher import AbstractPublisher

__all__ = ["RedisPublisher"]


class RedisPublisher(AbstractPublisher):
    """Implements a Redis publisher.

    Attributes:
        connection: Redis connection.

    """

    def __init__(self, **kwargs: Any) -> None:
        self._connection = redis.Redis(**kwargs)

    @property
    def connection(self) -> redis.Redis:
        """Redis connection."""
        return self._connection



# -*- coding: utf-8 -*-
# chemsys/adapters/listeners/redis_listener.py

# Standard Library Imports
from __future__ import annotations
import atexit
import logging
from typing import Any
from typing import Callable

# Third-Party Imports
import redis

# Local Imports
from .abstract_listener import AbstractListener

__all__ = ["RedisListener"]


# Initialize logger.
log = logging.getLogger("dodecahedron")


class RedisListener(AbstractListener):
    """Implements a Redis listener."""

    def __init__(self, **kwargs: Any) -> None:
        self._connection = redis.Redis(**kwargs)
        self._pubsub = self._connection.pubsub(ignore_subscribe_messages=True)  # type: ignore
        self._thread = None

    def start(self) -> None:
        """Start Redis listen."""
        self._thread = self._pubsub.run_in_thread(sleep_time=0.001)  # type: ignore
        atexit.register(self._thread.stop)

    def stop(self) -> None:
        """Stop Redis listener."""
        self._thread.stop()  # type: ignore
        atexit.unregister(self._thread.stop)  # type: ignore
        self._thread = None

    def subscribe(
        self, **kwargs: Callable[[str, Callable[..., Any]], None]
    ) -> None:
        """Subscribe event consumers.

        Args:
            **kwargs: Keyword arguments.

        """
        self._pubsub.subscribe(**kwargs)  # type: ignore



# -*- coding: utf-8 -*-

# Standard Library Imports
import collections
import functools
import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import overload
from typing import Type
from typing import Union

# Local Imports
from .messages import AbstractCommand
from .messages import AbstractEvent

__all__ = [
    "inject_handler_dependencies",
    "inject_dependencies",
    "merge_event_handlers",
]


# Custom types
Handler = Callable[..., None]
CommandHandlers = Dict[Type[AbstractCommand], Handler]
EventHandlers = Dict[Type[AbstractEvent], List[Handler]]


@overload
def inject_handler_dependencies(
    __handlers: CommandHandlers,
    /,
    dependencies: Dict[str, Any],
) -> CommandHandlers: ...


@overload
def inject_handler_dependencies(
    __handlers: EventHandlers,
    /,
    dependencies: Dict[str, Any],
) -> EventHandlers: ...


def inject_handler_dependencies(
    __handlers: Union[CommandHandlers, EventHandlers],
    /,
    dependencies: Dict[str, Any],
) -> Union[CommandHandlers, EventHandlers]:
    """Inject dependencies into handlers.

    Args:
        __handlers: Message handlers.
        dependencies: Dependencies.

    Returns:
        Handlers.

    """
    if all(issubclass(key, AbstractCommand) for key in __handlers.keys()):
        results = inject_command_handler_dependencies(__handlers, dependencies)  # type: ignore

    elif all(issubclass(key, AbstractEvent) for key in __handlers.keys()):
        results = inject_event_handler_dependencies(__handlers, dependencies)  # type: ignore

    else:
        raise NotImplementedError

    return results


def inject_command_handler_dependencies(
    __handlers: CommandHandlers, /, dependencies: Dict[str, Any]
) -> CommandHandlers:
    """Inject dependencies into command handlers.

    Based on 'Architecture Patterns in Python' dependency injection pattern.

    Args:
        __handlers: Command handlers.
        dependencies: Dependencies.

    Returns:
        Command handlers.

    Raises:
        TypeError: when `handlers` is not type 'dict'.

    .. _Architecture Patterns in Python:
        https://github.com/cosmicpython/code

    """
    if not isinstance(__handlers, dict):  # type: ignore
        message = f"expected type 'dict', got {type(__handlers)} instead"
        raise TypeError(message)

    results = {
        command_type: inject_dependencies(command_handler, dependencies)
        for command_type, command_handler in __handlers.items()
    }
    return results


def inject_event_handler_dependencies(
    __handlers: EventHandlers, /, dependencies: Dict[str, Any]
) -> EventHandlers:
    """Inject dependencies into event handlers.

    Based on 'Architecture Patterns in Python' dependency injection pattern.

    Args:
        __handlers: Event handlers.
        dependencies: Dependencies.

    Returns:
        Event handlers.

    Raises:
        TypeError: when `handlers` is not type 'dict'.

    .. _Architecture Patterns in Python:
        https://github.com/cosmicpython/code

    """
    if not isinstance(__handlers, dict):  # type: ignore
        message = f"expected type 'dict', got {type(__handlers)} instead"
        raise TypeError(message)

    results = {
        event_type: [
            inject_dependencies(handler, dependencies)
            for handler in event_handlers
        ]
        for event_type, event_handlers in __handlers.items()
    }
    return results


def inject_dependencies(
    __handler: Handler,
    /,
    dependencies: Dict[str, Any],
) -> Handler:
    """Inject dependencies into handler function.

    Based on 'Architecture Patterns in Python' dependency injection pattern.

    Args:
        __handler: Handler function.
        dependencies: Dependencies.

    .. _Architecture Patterns in Python:
        https://github.com/cosmicpython/code

    """
    params = inspect.signature(__handler).parameters
    kwargs = {
        name: dependency
        for name, dependency in dependencies.items()
        if name in params
    }
    result = functools.partial(__handler, **kwargs)
    return result


def merge_event_handlers(*args: EventHandlers) -> EventHandlers:
    """Merge event handlers.

    Args:
        *args: Groups of event handlers.

    Returns:
        Merged event handlers.

    """
    collection: EventHandlers = collections.defaultdict(list)
    for arg in args:
        for event_type, handlers in arg.items():
            collection[event_type].extend(handlers)

    result = dict(collection)
    return result



# -*- coding: utf-8 -*-
"""Abstract Mapper."""

# Standard Library Imports
from __future__ import annotations
import abc

__all__ = ["AbstractMapper"]


class AbstractMapper(abc.ABC):
    """Represents an abstract mapper."""



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib

# Local Imports
from .abstract_unit_of_work import *
from .eventful_unit_of_work import *
from .sessioned_unit_of_work import *

try:
    importlib.import_module("tqdm")  # pragma: no cover
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:
    from .progressive_unit_of_work import *  # pragma: no cover



# -*- coding: utf-8 -*-
"""Abstract Progress Bar."""

# Standard Library Imports
from __future__ import annotations
import abc
from typing import Optional
from typing import Union

__all__ = ["AbstractProgressBar"]


class AbstractProgressBar(abc.ABC):
    """Represents an abstract progress bar."""

    @property
    @abc.abstractmethod
    def current(self) -> int:
        """Current progress."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def leave(self) -> bool:
        """Whether progress bar is maintained between iterations."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def total(self) -> int:
        """Total progress."""
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """Close progress bar."""
        raise NotImplementedError

    @abc.abstractmethod
    def refresh(self) -> None:
        """Refresh progress bar."""
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self, total: Optional[Union[float, int]] = None) -> None:
        """Reset progress bar."""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self) -> None:
        """Update progress bar."""
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, message: str) -> None:
        """Write message to progress bar."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Class Mapper."""

# Standard Library Imports
from __future__ import annotations
from contextlib import contextmanager
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Generic
from typing import Hashable
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar

# Local Imports
from .abstract_mapper import AbstractMapper
from .schema import MapperSchema
from .schema import INWARD
from .schema import OUTWARD

__all__ = ["ClassMapper"]


# Custom types
T = TypeVar("T")


class ClassMapper(AbstractMapper, Generic[T]):
    """Implements a class mapper."""

    def __init__(
        self,
        __class: Type[T],
        /,
        schema: Dict[Hashable, Any],
        properties: Optional[Dict[Hashable, Any]] = None,
    ) -> None:
        super().__init__()
        self._class = __class
        self._schema = MapperSchema(schema)
        self._properties = properties

        setattr(self._class, "__mapper__", self)

    @property
    def cls(self) -> Type[T]:
        """Mapped class."""
        return self._class

    @property
    def schema(self) -> MapperSchema:
        """Mapped schema."""
        return self._schema

    @property
    def properties(self) -> Dict[Hashable, Any]:
        """Mapped properties."""
        results = self._properties or self._get_attribute_names()
        return results

    def from_dict(self, __dict: Dict[Hashable, Any], /) -> T:
        """Instantiate class from dictionary.

        Args:
            __dict: Dictionary.

        Returns:
            Instance of mapped class.

        """
        with replace_init(self.cls) as cls:
            result = cls()
            for attr, key in self.properties.items():
                converter = self._get_converter(key, direction=INWARD)
                value = converter(__dict[key]) if converter else __dict[key]
                setattr(result, str(attr), value)

        return result

    def from_list(self, __list: List[Any], /) -> T:
        """Instantiate class from list.

        Args:
            __list: List.

        Returns:
            Instance of mapped class.

        """
        with replace_init(self.cls) as cls:
            result = cls()
            for attr, idx in self.properties.items():
                converter = self._get_converter(idx, direction=INWARD)
                value = converter(__list[idx]) if converter else __list[idx]
                setattr(result, str(attr), value)

        return result

    def to_dict(self, __instance: T, /) -> Dict[Hashable, Any]:
        """Convert instance of class to dictionary.

        Args:
            __instance: Instance of mapped class.

        Returns:
            Dictionary.

        """
        result: Dict[Hashable, Any] = {}
        for attr, key in self.properties.items():
            converter = self._get_converter(key, direction=OUTWARD)
            value = getattr(__instance, str(attr))
            result[key] = converter(value) if converter else value

        return result

    def to_list(self, __instance: T, /) -> List[Any]:
        """Convert instance of class to list.

        Args:
            __instance: Instance of mapped class.

        Returns:
            List.

        """
        result: List[Any] = []
        for attr, key in sorted(self.properties.items(), key=lambda i: i[1]):
            converter = self._get_converter(key, direction=OUTWARD)
            value = getattr(__instance, str(attr))
            item = converter(value) if converter else value
            result.append(item)

        return result

    def _get_attribute_mapper(
        self, ref: Hashable
    ) -> Optional[Dict[Hashable, Any]]:
        """Get attribute mapper.

        Args:
            ref: Reference for attribute.

        Returns:
            Attribute mapper.

        """
        result = self.schema.get_attribute_mapper(ref)
        return result

    def _get_converter(
        self,
        ref: Hashable,
        direction: str,
        *,
        default: Callable[..., Any] = lambda x: x,  # type: ignore
    ) -> Optional[Callable[..., Any]]:
        """Get converter.

        Args:
            ref: Reference for attribute.
            direction: Direction of converter.
            default (optional): Default converter. Default `lambda x: x`.

        Returns:
            Converter.

        """
        result = self.schema.get_converter(ref, direction, default=default)
        return result

    def _get_attribute_names(self) -> Dict[Hashable, Hashable]:
        """Get attribute names.

        Returns:
            Attribute names.

        """
        results = self.schema.get_attribute_names()
        return results


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
@contextmanager
def replace_init(__class: type) -> Generator[type, None, None]:
    """Temporarily replace the ``__init__`` method on a class.

    Args:
        __class: Class.

    Yields:
        Class.

    """
    init = getattr(__class, "__init__")
    repl = getattr(object, "__init__")

    setattr(__class, "__init__", repl)
    yield __class
    setattr(__class, "__init__", init)



# -*- coding: utf-8 -*-
"""Abstract Wrapper."""

# Standard Library Imports
import abc
import logging
import os
import pathlib
from typing import Any
from typing import IO
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import TypeVar

# Local Imports
from .. import settings
from .. import utils

__all__ = [
    "AbstractFileSystemWrapper",
    "AbstractDirectoryWrapper",
    "AbstractFileWrapper",
]

# Initiate logger.
log = logging.getLogger("dodecahedron")

# Custom types
T = TypeVar("T")


class AbstractFileSystemWrapper(abc.ABC):
    """Represents an abstract file-system wrapper."""

    def __init__(self, *, read_only: bool = False) -> None:
        self.read_only = read_only

    @property
    @abc.abstractmethod
    def extension(self) -> str:
        """File extension."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def path(self) -> pathlib.Path:
        """File-system path."""
        raise NotImplementedError

    @property
    def read_only(self) -> bool:
        """Whether read only."""
        return self._read_only

    @read_only.setter
    def read_only(self, value: bool) -> None:
        if not isinstance(value, bool):  # type: ignore
            message = f"expected type 'bool', got {type(value)} instead"
            raise TypeError(message)

        self._read_only = value

    @abc.abstractmethod
    def open(self, *args: Any, **kwargs: Any) -> "IO[Any]":
        """Open file."""
        raise NotImplementedError

    def _open_file(
        self,
        filepath: "os.PathLike[Any]",
        mode: str,
    ) -> "IO[Any]":
        """Open a file and return a file object.

        Args:
            filepath: Path to file.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        self._raise_for_mode(mode)
        result = open(filepath, mode=mode)
        return result

    def _raise_for_mode(self, mode: str) -> None:
        """Raise for invalid mode.

        Args:
            mode: Mode.

        Raises:
            ValueError: when mode is invalid.

        """
        if self.read_only and set(mode).intersection({"a", "w", "x"}):
            message = f"{mode!s} mode not allowed when read-only is 'True'"
            raise ValueError(message)


class AbstractDirectoryWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for directories.

    Args:
        directory: Directory which contains file(s).
        read_only (optional): Whether directory is read only. Default ``False``.

    Raises:
        TypeError: when `directory` is not type ``PathLike``.
        NotADirectoryError: when `directory` is not a valid directory.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        extension: Optional[str] = None,
        read_only: bool = False,
    ) -> None:
        if not isinstance(directory, os.PathLike):  # type: ignore
            expected = "expected type 'PathLike'"
            actual = f"got {type(directory)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        if extension and not isinstance(extension, str):  # type: ignore
            message = f"expected type 'str', got {type(extension)} instead"
            raise TypeError(message)

        if not os.path.exists(directory) or not os.path.isdir(directory):
            message = f"{directory!s} is not a valid directory"
            raise NotADirectoryError(message)

        super().__init__(read_only=read_only)
        self._directory = pathlib.Path(directory)
        self._extension = utils.standardize_file_extension(extension or "*")

    @property
    def path(self) -> pathlib.Path:
        """Path to directory."""
        return self.directory

    @property
    def directory(self) -> pathlib.Path:
        """Path to directory."""
        return self._directory

    @property
    def extension(self) -> str:
        """File extension."""
        return self._extension

    def __fspath__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return str(self.directory)

    def open(self, filename: str, /, mode: str = "r") -> "IO[Any]":
        """Open a file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """
        filepath = self._get_filepath(filename)
        result = self._open_file(filepath, mode)
        return result

    def _get_filepath(self, filename: str) -> pathlib.Path:
        """Get filepath.

        Args:
            filename: Filename.

        Returns:
            Filepath.

        """
        result = self._directory / filename
        if not result.exists():
            result = self.find(filename)

        utils.raise_for_extension(result, self.extension)
        return result

    def find(self, ref: str) -> pathlib.Path:
        """Find path for file in directory.

        Finds the filepath for a file in the source directory where the
        filename contains the provided substring.

        Args:
            ref: Substring for which to search.

        Returns:
            Path for file.

        Raises:
            FileNotFoundError: When no filenames match provided substring.

        """
        log.debug(
            "Searching for %(ref)s in %(dir)s",
            {"ref": ref, "dir": self._directory},
        )
        filename = utils.set_extension(ref, self.extension)
        result = self._find_file_in_directory(filename)
        log.debug(
            "Found %(file)s in %(dir)s",
            {"file": result, "dir": self._directory},
        )
        return result

    def _find_file_in_directory(self, __filename: str) -> pathlib.Path:
        """Find path for file in directory.

        Finds the filepath for a file in the source directory where the
        filename contains the provided substring.

        Args:
            __filename: Name of file.

        Returns:
            Path for file.

        Raises:
            FileNotFoundError: When no files match provided filename.

        """
        try:
            filepath = next(path for path in self._directory.rglob(__filename))

        except StopIteration as err:
            message = f"{self._directory / __filename} not found"
            raise FileNotFoundError(message) from err

        return filepath


class AbstractFileWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for files.

    Args:
        filepath: Path to file.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        TypeError: when `filepath` is not type ``PathLike``.
        IsADirectoryError: when `filepath` points to a directory.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        read_only: bool = False,
    ) -> None:
        if not isinstance(filepath, os.PathLike):  # type: ignore
            expected = "expected type 'PathLike'"
            actual = f"got {type(filepath)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        if os.path.isdir(filepath):
            message = f"{filepath!s} is a directory"
            raise IsADirectoryError(message)

        super().__init__(read_only=read_only)
        self._filepath = pathlib.Path(filepath)

    @property
    def path(self) -> pathlib.Path:
        """Path to file."""
        return self.filepath

    @property
    def filepath(self) -> pathlib.Path:
        """Path to file."""
        return self._filepath

    @property
    def extension(self) -> str:
        """File extension."""
        return self._filepath.suffix

    def __fspath__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return str(self.filepath)

    def open(self, mode: str = "rb") -> "IO[Any]":
        """Open file and return a file object.

        Args:
            mode (optional): Mode. Default ``rb``.

        Returns:
            File object.

        """
        result = self._open_file(self._filepath, mode=mode)
        return result


class AbstractTextWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for text files.

    Args:
        *args: Positional arguments.
        encoding (optional): File encoding. Default `utf-8`.
        newline (optional): Newline character. Default ``None``.
        **kargs: Keyword arguments.

    Raises:
        TypeError: when `encoding` is not type ``str``.

    """

    def __init__(
        self,
        *args: Any,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        newline: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        if not isinstance(encoding, str):  # type: ignore
            message = f"expected type 'str', got {type(encoding)} instead"
            raise TypeError(message)

        self._encoding = encoding
        self._newline = newline

    @property
    def encoding(self) -> str:
        """Expected file encoding."""
        return self._encoding

    @property
    def newline(self) -> Optional[str]:
        """Newline character."""
        return self._newline

    def _open_file(
        self,
        filepath: "os.PathLike[Any]",
        mode: str,
    ) -> IO[Any]:
        """Open a file and return a file object.

        Args:
            filepath: Path to file.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        encoding = self._get_encoding_for_file_mode(mode)
        newline = self._get_newline_for_file_mode(mode)
        result = open(
            filepath,
            mode=mode,
            encoding=encoding,
            newline=newline,
        )
        return result

    def _get_encoding_for_file_mode(self, mode: str) -> Optional[str]:
        """Get encoding for file mode.

        Args:
            mode: Mode.

        Returns:
            Encoding.

        """
        result = self._encoding if "b" not in mode else None
        return result

    def _get_newline_for_file_mode(self, mode: str) -> Optional[str]:
        """Get newline for file mode.

        Args:
            mode: Mode.

        Returns:
            Newline.

        """
        result = self._newline if "b" not in mode else None
        return result


class AbstractIOWrapper(IO[Any]):  # type: ignore
    """Represents an abstract I/O wrapper class for files.

    Args:
        __file: File.

    """

    @property
    @abc.abstractmethod
    def file(self) -> IO[Any]:
        """File."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        """Whether file is closed."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def read_only(self) -> bool:
        """Whether read only."""
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """Close `.xlsx` file."""
        raise NotImplementedError

    def __iter__(self) -> Iterator[Any]:
        return self.file.__iter__()

    def __next__(self) -> Any:
        return self.file.__next__()

    def fileno(self) -> int:
        return self.file.fileno()

    def flush(self) -> None:
        return self.file.flush()

    def isatty(self) -> bool:
        return self.file.isatty()

    def read(self, n: int = -1) -> Any:
        return self.file.read(n)

    def readable(self) -> bool:
        return self.file.readable()

    def readline(self, limit: int = -1) -> Any:
        return self.file.readline(limit)

    def readlines(self, hint: int = -1) -> list[Any]:
        return self.file.readlines(hint)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.file.seek(offset, whence)

    def seekable(self) -> bool:
        return self.file.seekable()

    def tell(self) -> int:
        return self.file.tell()

    def truncate(self, size: Optional[int] = None) -> int:
        return self.file.truncate(size)

    def write(self, s: Any) -> int:
        return self.file.write(s)

    def writable(self) -> bool:
        return self.file.writable()

    def writelines(self, lines: Iterable[Any], /) -> None:
        return self.file.writelines(lines)



# -*- coding: utf-8 -*-

# Standard Library Imports
from __future__ import annotations

# Third-Party Imports
from ..units_of_work import AbstractUnitOfWork
from ..units_of_work import EventfulUnitOfWork

__all__ = ["FakeUnitOfWork", "FakeEventfulUnitOfWork"]


# Private Attributes
COMMITTED_ATTR = "_committed"
ROLLED_BACK_ATTR = "_rolled_back"


class FakeUnitOfWork(AbstractUnitOfWork):
    """Implements a fake unit of work."""

    @property
    def committed(self) -> bool:
        """Whether `commit()` method was called."""
        result = getattr(self, COMMITTED_ATTR, False)
        return result

    @property
    def rolled_back(self) -> bool:
        """Whether `rollback()` method was called."""
        result = getattr(self, ROLLED_BACK_ATTR, False)
        return result

    def __enter__(self) -> FakeUnitOfWork:
        return self

    def commit(self) -> None:
        """Commit changes."""
        setattr(self, COMMITTED_ATTR, True)

    def rollback(self) -> None:
        """Rollback changes."""
        setattr(self, ROLLED_BACK_ATTR, True)


class FakeEventfulUnitOfWork(FakeUnitOfWork, EventfulUnitOfWork):
    """Implements a fake eventful unit of work."""



# -*- coding: utf-8 -*-
"""Float Converter.

Module provides functions for converting values to floats.

"""

# Standard Library Imports
import datetime
import decimal
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional

# Local Imports
from .base_converter import BaseConverter
from .. import parsers

__all__ = ["to_float"]


def to_float(__value: Any, /, default: float = 0.0) -> float:
    """Convert value to float.

    Args:
        __value: Value to convert to float.
        default (optional): Default value. Default ``0.0``.

    Returns:
        Float.

    """
    converter = FloatConverter(default=default)
    result = converter(__value)
    return result


class FloatConverter(BaseConverter):
    """Class implements a float converter.

    Args:
        default (optional): Default value. Default ``0.0``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: float = 0.0,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, float):
            message = f"expected type 'float', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()

    @property
    def default(self) -> Any:  # pragma: no cover
        """Default value."""
        return self._default

    @default.setter
    def default(self, value: Any) -> None:  # pragma: no cover
        if not isinstance(value, float):  # type: ignore
            message = f"expected type 'float', got {type(value)} instead"
            raise TypeError(message)

        self._default = value


def float_from_bool(__value: bool, _: float, /) -> float:
    """Convert boolean value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'bool'.

    """
    if not isinstance(__value, bool):  # type: ignore  # pragma: no cover
        message = f"expected type 'bool', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def float_from_date(
    __value: datetime.date, _: Optional[float] = None, /
) -> float:
    """Convert date value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore  # pragma: no cover
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    result = time.mktime(__value.timetuple())
    return result


def float_from_datetime(
    __value: datetime.datetime, _: Optional[float] = None, /
) -> float:
    """Convert datetime value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore  # pragma: no cover
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = __value.timestamp()
    return result


def float_from_decimal(__value: decimal.Decimal, _: float, /) -> float:
    """Convert decimal value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore  # pragma: no cover
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def float_from_float(__value: float, _: float, /) -> float:
    """Convert float value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):  # type: ignore  # pragma: no cover
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def float_from_int(__value: int, _: float, /) -> float:
    """Convert integer value to float.

    Args:
        __value: Value to convert to float.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore  # pragma: no cover
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def float_from_str(__value: str, default: float = 0.0, /) -> float:
    """Convert string value to float.

    Args:
        __value: String representation of float value.

    Returns:
        Float.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to float.

    """
    if not isinstance(__value, str):  # type: ignore  # pragma: no cover
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    try:
        result = parsers.parse_number(value)

    except ValueError:
        message = f"{type(__value)} cannot be converted to float"
        raise ValueError(message)

    else:
        return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., float]] = {
    bool: float_from_bool,
    datetime.date: float_from_date,
    datetime.datetime: float_from_datetime,
    decimal.Decimal: float_from_decimal,
    float: float_from_float,
    int: float_from_int,
    str: float_from_str,
}



# -*- coding: utf-8 -*-

# Standard Library Imports
import abc
from dataclasses import asdict
from dataclasses import is_dataclass
import logging
import json
from typing import Any

# Local Imports
from ..json import JSONEncoder
from ..messages import event

__all__ = ["AbstractPublisher"]


# Initialize logger.
log = logging.getLogger("dodecahedron")


class AbstractPublisher(abc.ABC):
    """Represents an abstract publisher."""

    @property
    @abc.abstractmethod
    def connection(self) -> Any:
        """Connection."""
        raise NotImplementedError

    def publish(self, channel: str, event: event.AbstractEvent, /) -> None:
        """publishes an event to an external message broker.

        Args:
            channel: Channel on which to publish event.
            event: Event to publish on external broker.

        """
        log.info("publishing: channel=%s, event=%s", channel, event)
        data = asdict(event) if is_dataclass(event) else event
        payload = json.dumps(data, cls=JSONEncoder)
        getattr(self.connection, "publish")(channel, payload)



# -*- coding: utf-8 -*-

# Standard Library Imports
import abc
from typing import Any
from types import MethodType

# Local Imports
from .abstract_repository import AbstractRepository

__all__ = ["SessionedRepository"]


class SessionedRepository(AbstractRepository):
    """Represents a sessioned repository.

    Attributes:
        session: Session.

    """

    @property
    @abc.abstractmethod
    def session(self) -> Any:
        """Session."""
        raise NotImplementedError

    def close(self, *args: Any, **kwargs: Any) -> Any:
        """Close connection to repository.

        Args:
            *args (optional): Positional arguments.
            **kwargs (optional): Keyword arguments.

        """
        result = self._call_method_on_session("close", *args, **kwargs)
        return result

    def commit(self, *args: Any, **kwargs: Any) -> Any:
        """Commit changes to repository.

        Args:
            *args (optional): Positional arguments.
            **kwargs (optional): Keyword arguments.

        """
        result = self._call_method_on_session("commit", *args, **kwargs)
        return result

    def rollback(self, *args: Any, **kwargs: Any) -> Any:
        """Rollback changes to repository.

        Args:
            *args (optional): Positional arguments.
            **kwargs (optional): Keyword arguments.

        """
        result = self._call_method_on_session("rollback", *args, **kwargs)
        return result

    def _call_method_on_session(
        self, __name: str, /, *args: Any, **kwargs: Any
    ) -> Any:
        """Call method on session.

        Args:
            __name: Name of method.
            *args (optional): Positional arguments.
            **kwargs (optional): Keyword arguments.

        """
        try:
            method: MethodType = getattr(self.session, __name)
            result = method(*args, **kwargs)

        except AttributeError:
            return None

        return result



# -*- coding: utf-8 -*-

# Standard Library Imports
from __future__ import annotations
import abc
import inspect
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

__all__ = ["SingletonMeta"]


# Custom types
T = TypeVar("T")


class SingletonMeta(abc.ABCMeta):
    """Implements a singleton metaclass.

    The singleton metaclass allows subclasses to act as singletons.

    By implementing the singleton pattern using a metaclass, instead of within
    the `__new__` method, we avoid calling `__init__` each time a subclass is
    instantiated.

    Notes:
        * Subclasses will only behave as singletons when they contain a
        `__singleton__` attribute which is set to ``True``.

    """

    __instances__: Dict[str, Any] = {}

    def __call__(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        key = make_key(cls)

        if key not in SingletonMeta.__instances__:
            instance = super().__call__(*args, **kwargs)
            if SingletonMeta.is_singleton(instance):
                SingletonMeta.__instances__[key] = instance
        else:
            instance = SingletonMeta.__instances__[key]

        return instance

    @staticmethod
    def clear() -> None:
        """Clear all instances of subclasses."""
        SingletonMeta.__instances__.clear()

    @staticmethod
    def discard(__subclass: Union[object, type]) -> None:
        """Discard an instance of a subclass.

        Args:
            __subclass: Subclass to discard.

        """
        subclass = get_class(__subclass)
        key = make_key(subclass)
        if SingletonMeta.__instances__.get(key):
            del SingletonMeta.__instances__[key]

    @staticmethod
    def is_singleton(__obj: Union[object, type], /) -> bool:
        """Checks whether an object is a singleton.

        Args:
            __obj: Object.

        Returns:
            Whether object is a singleton.

        """
        result: bool = getattr(__obj, "__singleton__", False)
        return result


def make_key(__cls: type, /) -> str:
    """Make a unique key for a class.

    Args:
        __cls: Class for which to make key.

    Returns:
        Key.

    """
    result = "%s.%s" % (__cls.__module__, __cls.__name__)
    return result


def get_class(__class_or_object: Any, /) -> type:
    """Get class of object.

    Args:
        __class_or_object: Class or instance.

    Returns:
        Class.

    """
    result = (
        # Unwrap argument in case class is decorated.
        inspect.unwrap(__class_or_object).__class__
        if not inspect.isclass(__class_or_object)
        else __class_or_object
    )
    return result



# -*- coding: utf-8 -*-
"""Message Broker."""

# Standard Library Imports
from __future__ import annotations
import abc
from collections import defaultdict
from datetime import datetime
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

# Local Imports
from .metaclasses import SingletonMeta
from . import config

__all__ = ["MessageBroker"]


# Initialize logger.
log = logging.getLogger("dodecahedron")

# Custom types
Subscriber = Callable[[Dict[str, Any]], None]

# Constants
IGNORE = "ignore"
RAISE = "raise"


class AbstractMessageBroker(abc.ABC):
    """Represents an abstract message broker."""

    @property
    @abc.abstractmethod
    def channels(self) -> List[str]:
        """Channels."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def subscribers(self) -> Dict[str, List[Subscriber]]:
        """Subscribers."""
        raise NotImplementedError

    @abc.abstractmethod
    def publish(self, channel: str, event: str) -> None:
        """Publish an event to a channel.

        Args:
            channel: Name of channel.
            event: JSON representation of an event.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def subscribe(self, channel: str, subscriber: Subscriber) -> None:
        """Add subscriber to channel.

        Args:
            channel: Name of channel to which to subscribe.
            subscriber: Function to call for events on channel.

        """
        raise NotImplementedError


class MessageBroker(AbstractMessageBroker, metaclass=SingletonMeta):
    """Implements a message broker."""

    __singleton__ = True

    def __init__(self) -> None:
        self._subscribers: Dict[str, List[Subscriber]] = defaultdict(list)

    @property
    def channels(self) -> List[str]:
        """Channels."""
        return list(self._subscribers.keys())

    @property
    def subscribers(self) -> Dict[str, List[Subscriber]]:
        """Subscribers."""
        return self._subscribers

    def publish(self, channel: str, event: str) -> None:
        """Publish an event to a channel.

        Args:
            channel: Name of channel.
            event: JSON representation of an event.

        """
        message = self.make_message(event)
        self.send_message(channel, message)

    @staticmethod
    def make_message(event: str, /) -> Dict[str, Any]:
        """Make message to send to subscribers.

        Args:
            event: Published event.

        Returns:
            Message.

        """
        message: Dict[str, Any] = {"data": event, "created_at": datetime.now()}
        return message

    def send_message(
        self, channel: str, message: Dict[str, Any]
    ) -> None:  # pylint: disable=broad-except
        """Send message to channel subscribers.

        Args:
            channel: Name of channel.
            message: Message to send to subscribers.

        """
        for subscriber in self.subscribers[channel]:
            try:
                log.debug(
                    "sending %(event)s event to subscriber %(subscriber)s",
                    {"event": channel, "subscriber": subscriber},
                )
                subscriber(message)
            except Exception as error:
                handle_error(channel, error, on_error=IGNORE)
                continue

    def subscribe(self, channel: str, subscriber: Subscriber) -> None:
        """Add subscriber to channel.

        Args:
            channel: Name of channel to which to subscribe.
            subscriber: Function to call for events on channel.

        """
        self.subscribers[channel].append(subscriber)


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def handle_error(
    event: str,
    error: Exception,
    *,
    on_error: Optional[Literal["ignore", "raise"]],
) -> None:
    """Handle error raised when handling event.

    Args:
        event: Event.
        error: Error raised when handling event.
        on_error: Strategy for handling errors.

    """
    exc_info = not config.is_production_environment()
    log.error("Error handling %s", event, exc_info=exc_info)
    if on_error == RAISE:
        raise error



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib

# Local Imports
from .abstract_listener import *

try:
    importlib.import_module("redis")  # pragma: no cover
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:  # pragma: no cover
    from .redis_listener import *



# -*- coding: utf-8 -*-

# Local Imports
from .singleton import SingletonMeta
from .tracker import TrackerMeta

__all__ = ["RepositoryMeta"]


class RepositoryMeta(SingletonMeta, TrackerMeta):
    """Implements a repository metaclass.

    Combines the singleton and tracker metaclasses into a single metaclass
    for use by repositories.

    """



# -*- coding: utf-8 -*-
"""Yes or No Converter.

Module provides function for converting values to 'Yes' or 'No'.

"""

# Standard Library Imports
from typing import Literal
from typing import Optional

# Local Imports
from .boolean_converter import BooleanConverter

__all__ = ["to_y_or_n", "to_yes_or_no"]


def to_y_or_n(
    __value: object, /, default: Optional[str] = None
) -> Literal["Y", "N"]:
    """Convert value to `Y` or `N`.

    Args:
        value: Value to convert.
        default (optional): Default value. Default ``None``.

    Returns:
        `Y` or `N`.

    """
    if default and default not in ("Y", "N"):
        message = f"expected value of 'Y' or 'N', got {default} instead"
        raise ValueError(message)

    try:
        converter = BooleanConverter()
        converter.default = converter(default)
        value = converter(__value)

    except TypeError:
        message = f"{type(__value)} cannot be converted to 'Y' or 'N'"
        raise TypeError(message)

    except ValueError:
        message = f"'{__value}' cannot be converted to 'Y' or 'N'"
        raise ValueError(message)

    result: Literal["Y", "N"] = "Y" if value is True else "N"
    return result


def to_yes_or_no(
    __value: object, /, default: Optional[str] = None
) -> Literal["Yes", "No"]:
    """Convert value to `Yes` or `No`.

    Args:
        value: Value to convert.
        default (optional): Default value. Default ``None``.

    Returns:
        `Yes` or `No`.

    """
    if default and default not in ("Yes", "No"):
        message = f"expected value of 'Yes' or 'No', got {default} instead"
        raise ValueError(message)

    try:
        converter = BooleanConverter()
        converter.default = converter(default)
        value = converter(__value)

    except TypeError as error:
        message = f"{type(__value)} cannot be converted to 'Yes' or 'No'"
        raise TypeError(message) from error

    except ValueError as error:
        message = f"'{__value}' cannot be converted to 'Yes' or 'No'"
        raise ValueError(message) from error

    result: Literal["Yes", "No"] = "Yes" if value is True else "No"
    return result



# -*- coding: utf-8 -*-

# Standard Library Imports
from typing import Any
import warnings

# Third-Party Imports
from sqlalchemy.orm import Session

# Local Imports
from .sessioned_repository import SessionedRepository
from sqlalchemy import exc as sa_exc

__all__ = ["SqlAlchemyRepository"]


class SqlAlchemyRepository(SessionedRepository):
    """Class implements an SQLAlchemy repository.

    The repository uses SQLAlchemy to read data from a database and to handle
    relevant CRUD operations.

    Attributes:
        session: SQLAlchemy session.
        *args (optional): Positional arguments.
        **kwargs (optional): Keyword arguments.

    .. _SQLAlchemy Documentation:
        https://docs.sqlalchemy.org/


    """

    def __init__(self, session: Session, /, *args: Any, **kwargs: Any) -> None:
        if not isinstance(session, Session):  # type: ignore
            message = f"expected type 'Session', got {type(session)} instead"
            raise TypeError(message)

        super().__init__(*args, **kwargs)
        self._session = session

    @property
    def session(self) -> Session:
        """Session."""
        return self._session

    def commit(self, *args: Any, **kwargs: Any) -> Any:
        """Commit changes to repository.

        Args:
            *args (optional): Positional arguments.
            **kwargs (optional): Keyword arguments.

        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)
            result = super().commit(*args, **kwargs)

        return result

    def execute(self, *args: Any, **kwargs: Any) -> Any:
        """Call the execute method directly on the SQLAlchemy session.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            Result.

        """
        result = self.session.execute(*args, **kwargs)
        return result



# -*- coding: utf-8 -*-
"""PDF File Wrappers."""

# Standard Library Imports
from __future__ import annotations
import os
from typing import Any
from typing import IO
from typing import Optional

# Local Imports
from .abstract_file_wrappers import AbstractDirectoryWrapper
from .abstract_file_wrappers import AbstractFileSystemWrapper
from .abstract_file_wrappers import AbstractFileWrapper
from .abstract_file_wrappers import AbstractIOWrapper
from ..utils import converters
from .. import settings
from .. import utils

__all__ = ["PdfDirectoryWrapper", "PdfFileWrapper"]


class AbstractPdfWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for `.pdf` files."""

    def _init_pdf_io_wrapper(self, __file: IO[Any], /) -> PdfIOWrapper:
        """Initialize I/O wrapper for `.pdf` file.

        Args:
            __file: File-like object.

        Returns:
            I/O wrapper instance.

        """
        result = PdfIOWrapper(__file)
        setattr(result, "_context", self)
        return result


class PdfDirectoryWrapper(AbstractPdfWrapper, AbstractDirectoryWrapper):
    """Implements a wrapper for `.pdf` files in a directory.

    Args:
        directory: Directory from which to load `.pdf` file(s).
        read_only (optional): Whether file is read only. Default ``False``.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            directory,
            extension=settings.PDF_EXTENSION,
            read_only=read_only,
        )

    def open(self, filename: str, /, mode: str = "rb") -> IO[Any]:
        """Open a `.pdf` file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``rb``.

        Returns:
            File object.

        """
        file = super().open(filename, mode=converters.to_bytes_file_mode(mode))
        result = self._init_pdf_io_wrapper(file)
        return result


class PdfFileWrapper(AbstractPdfWrapper, AbstractFileWrapper):
    """Implements a wrapper for `.pdf` files.

    Args:
        filepath: Path to `.pdf` file.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        ValueError: when `filepath` is not a `.pdf` file.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        read_only: bool = False,
    ) -> None:
        super().__init__(filepath, read_only=read_only)
        utils.raise_for_extension(filepath, settings.PDF_EXTENSION)

    def open(self, mode: str = "rb") -> IO[Any]:
        """Open the `.pdf` file and return a file object.

        Args:
            mode (optional): Mode. Default ``rb``.

        Returns:
            File object.

        """
        file = super().open(converters.to_bytes_file_mode(mode))
        result = self._init_pdf_io_wrapper(file)
        return result


class PdfIOWrapper(AbstractIOWrapper):
    """Implements a I/O wrapper for `.pdf` files."""

    def __init__(self, __file: IO[Any]) -> None:
        self._file = __file
        self._context: Optional[AbstractPdfWrapper] = None

    @property
    def file(self) -> IO[Any]:
        """File."""
        return self._file

    @property
    def closed(self) -> bool:
        """Check whether file is closed.

        Returns:
            Whether file is closed.

        """
        return self._file.closed

    def __enter__(self) -> PdfIOWrapper:
        return self

    def __exit__(self, *_) -> None:
        self.close()
        return

    def close(self) -> None:
        """Close `.pdf` file."""
        self._file.close()



# -*- coding: utf-8 -*-

# Standard Library Imports
import functools
from typing import Any
from typing import Hashable
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import Optional
from typing import Sequence
from typing import Union

__all__ = [
    "deep_get",
    "deep_set",
]


def deep_get(
    __m: Union[Mapping[Hashable, Any], Sequence[Any]],
    /,
    key: Union[str, Sequence[str]],
    default: Optional[object] = None,
) -> Any:
    """Get value from nested mapping object.

    Args:
        __d: Mapping object.
        key: Key(s) seperated by periods or in a sequence.
        default (optional): Default when value not found. Default ``None``.

    Returns:
        Value of key in nested mapping object.

    Raises:
        TypeError: when argument is not a mapping.

    """

    def _get_item(o: Any, k: str) -> Any:
        """Get item from object.

        Args:
            o: Object.
            key: Key for which to get value.

        Returns:
            Value.

        """
        try:
            result = o[k] if not k.isdigit() else o[int(k)]

        except (IndexError, KeyError):
            return o[_standardize_key(k)]

        return result

    def _get(o: Any, k: str) -> Any:
        """Get value of key from object.

        Args:
            o: Object.
            k: Key for which to get value.

        Returns:
            Value.

        """
        try:
            result = _get_item(o, k)

        except (IndexError, KeyError, TypeError):
            return default

        return result

    if not isinstance(__m, (Mapping, Sequence)):  # type: ignore
        message = f"expected mapping or sequence, got {type(__m)} instead"
        raise TypeError(message)

    if not isinstance(key, (str, Sequence)):  # type: ignore
        message = f"expected type 'str', got {type(key)} instead"
        raise TypeError(message)

    keys: List[str] = key.split(".") if isinstance(key, str) else list(key)
    result = functools.reduce(_get, keys, __m)
    return result


def deep_set(
    __m: Union[MutableMapping[Hashable, Any], Sequence[Any]],
    /,
    key: Union[str, Sequence[str]],
    value: Any,
) -> Union[MutableMapping[Hashable, Any], Sequence[Any]]:
    """Sets key to value in nested mapping object.

    Args:
        __m: Mapping.
        key: Key(s) seperated by periods or a sequence of keys.
        value: Value to set for key.

    Returns:
        Updated mutable mapping object.

    Raises:
        TypeError: when argument is not mutable mapping.

    """

    def _set_item(o: Any, k: str, v: Any) -> None:
        """Set value of key on object.

        Args:
            o: Object.
            k: Key.
            v: Value.

        """
        if isinstance(o, dict):
            o[k] = v

        if isinstance(o, list) and k.isdigit():
            o[int(k)] = v

        if isinstance(o, list) and not k.isdigit():
            raise KeyError(k)

        if not isinstance(o, (dict, list)):
            raise KeyError(k)

    if not isinstance(__m, (MutableMapping, Sequence)):  # type: ignore
        expected = "expected mutable mapping or sequence"
        actual = f"got {type(__m)} instead"
        message = ", ".join([expected, actual])
        raise TypeError(message)

    keys: List[str] = key.split(".") if isinstance(key, str) else list(key)

    try:
        k, remaining = keys[0], keys[1:]
        v = (
            deep_set(deep_get(__m, k) or {}, remaining, value)
            if len(remaining) >= 1
            else value
        )
        _set_item(__m, k, v)

    except KeyError as error:
        raise KeyError(f"{k}.{error.args[0]}") from error  # type: ignore

    except (IndexError, TypeError) as error:
        raise KeyError(keys[0]) from error

    return __m


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def _standardize_key(__k: str, /) -> str:
    """Standardize key.

    Args:
        k: Key.

    Returns:
        Key.

    """
    if __k.startswith("'") and __k.endswith("'"):
        return __k.strip("'")

    if __k.startswith('"') and __k.endswith('"'):
        return __k.strip('"')

    if __k.startswith("[") and __k.endswith("]"):
        return __k.strip("[").strip("]")

    if __k.startswith("(") and __k.endswith(")"):
        return __k.strip("(").strip(")")

    if __k.startswith("<") and __k.endswith(">"):
        return __k.strip("<").strip(">")

    return __k



# -*- coding: utf-8 -*-
"""File-System Repository."""

# Standard Library Imports
import pathlib
from typing import Any
from typing import Optional

# Local Imports
from .abstract_repository import AbstractRepository
from ..wrappers import AbstractFileSystemWrapper

__all__ = ["AbstractFileSystemRepository"]


class AbstractFileSystemRepository(AbstractRepository):
    """Represents an abstract file-system repository.

    Args:
        wrapper: File-system wrapper.
        *args (optional): Positional arguments.
        **kwargs (optional): Keyword arguments.

    """

    def __init__(self, wrapper: Any, /, *args: Any, **kwargs: Any) -> None:
        if not isinstance(wrapper, AbstractFileSystemWrapper):
            expected = "expected type 'AbstractFileSystemWrapper'"
            actual = f"got {type(wrapper)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(*args, **kwargs)
        self._wrapper = wrapper

    @property
    def path(self) -> Optional[pathlib.Path]:
        """Path."""
        result = getattr(self._wrapper, "path", None)
        return result



# -*- coding: utf-8 -*-
"""Progressive Dispatcherr."""

# Local Imports
from .abstract_dispatcher import AbstractDispatcher
from ..messagebus import AbstractMessageBus
from ..progress import AbstractProgressBar
from ..units_of_work import ProgressiveUnitOfWork

__all__ = ["ProgressiveDispatcher"]


class ProgressiveDispatcher(AbstractDispatcher):
    """Implements a progressive dispatcher.

    Args:
        __bus: Message bus.

    Attributes:
        messagebus: Message bus.
        progress: Progress bar.

    """

    def __init__(self, __bus: AbstractMessageBus, /) -> None:
        if not isinstance(__bus.uow, ProgressiveUnitOfWork):
            expected = "expected type 'ProgressiveUnitOfWork'"
            actual = f"got {type(__bus.uow)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(__bus)

    @property
    def progress(self) -> AbstractProgressBar:
        """Progress bar."""
        result = getattr(self.messagebus.uow, "progress")
        return result



# -*- coding: utf-8 -*-
"""Dictionary Mapper."""

# Standard Library Imports
from __future__ import annotations
from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import List
from typing import Optional

# Local Imports
from .abstract_mapper import AbstractMapper
from .schema import MapperSchema
from .schema import INWARD
from .schema import OUTWARD

__all__ = ["DictMapper"]


class DictMapper(AbstractMapper):
    """Implements a dictionary mapper."""

    def __init__(
        self,
        schema: Dict[Hashable, Any],
        properties: Optional[Dict[Hashable, Any]] = None,
    ) -> None:
        super().__init__()
        self._schema = MapperSchema(schema)
        self._properties = properties

    @property
    def schema(self) -> MapperSchema:
        """Mapped schema."""
        return self._schema

    @property
    def properties(self) -> Dict[Hashable, Any]:
        """Mapped properties."""
        results = self._properties or self._get_attribute_names()
        return results

    def from_dict(self, __dict: Dict[Hashable, Any], /) -> Dict[Hashable, Any]:
        """Instantiate dictionary from dictionary.

        Args:
            __dict: Dictionary.

        Returns:
            Dictionary.

        """
        if not isinstance(__dict, dict):  # type: ignore
            message = f"expected type 'dict', got {type(__dict)} instead"
            raise TypeError(message)

        result: Dict[Hashable, Any] = {}
        for outward_key, inward_key in self.properties.items():
            converter = self._get_converter(inward_key, direction=INWARD)
            value: Any = __dict[outward_key]
            result[str(inward_key)] = converter(value) if converter else value

        return result

    def from_list(self, __list: List[Any], /) -> Dict[Hashable, Any]:
        """Instantiate dictionary from list.

        Args:
            __list: List.

        Returns:
            Dictionary.

        """
        result: Dict[Hashable, Any] = {}
        for key, idx in self.properties.items():
            converter = self._get_converter(idx, direction=INWARD)
            value: Any = __list[idx]
            result[key] = converter(value) if converter else value

        return result

    def to_dict(self, __dict: Dict[Hashable, Any], /) -> Dict[Hashable, Any]:
        """Convert instance of class to dictionary.

        Args:
            __dict: Dictionary.

        Returns:
            Dictionary.

        """
        result: Dict[Hashable, Any] = {}
        for outward_key, inward_key in self.properties.items():
            converter = self._get_converter(inward_key, direction=OUTWARD)
            value: Any = __dict[inward_key]
            result[str(outward_key)] = converter(value) if converter else value

        return result

    def to_list(self, __dict: Dict[Hashable, Any], /) -> List[Any]:
        """Convert instance of class to list.

        Args:
            __dict: Dictionary.

        Returns:
            List.

        """
        result: List[Any] = []
        for key, idx in sorted(self.properties.items(), key=lambda i: i[1]):
            converter = self._get_converter(idx, direction=OUTWARD)
            value: Any = __dict[key]
            item = converter(value) if converter else value
            result.append(item)

        return result

    def _get_attribute_mapper(
        self, ref: Hashable
    ) -> Optional[Dict[Hashable, Any]]:
        """Get attribute mapper.

        Args:
            ref: Reference for attribute.

        Returns:
            Attribute mapper.

        """
        result = self.schema.get_attribute_mapper(ref)
        return result

    def _get_converter(
        self,
        ref: Hashable,
        direction: str,
        *,
        default: Callable[..., Any] = lambda x: x,  # type: ignore
    ) -> Optional[Callable[..., Any]]:
        """Get converter.

        Args:
            ref: Reference for attribute.
            direction: Direction of converter.
            default (optional): Default converter. Default `lambda x: x`.

        Returns:
            Converter.

        """
        result = self.schema.get_converter(ref, direction, default=default)
        return result

    def _get_attribute_names(self) -> Dict[Hashable, Hashable]:
        """Get attribute names.

        Returns:
            Attribute names.

        """
        results = self.schema.get_attribute_names()
        return results



# -*- coding: utf-8 -*-
"""Xlsx Repository."""

# Standard Library Imports
import logging
from typing import Any
from typing import Dict
from typing import Hashable
from typing import List
from typing import Optional
from typing import Sequence

# Local Imports
from .file_system_repository import AbstractFileSystemRepository
from ..wrappers.xlsx_file_wrappers import XlsxFileWrapper
from ..mappers import ClassMapper

__all__ = ["AbstractXlsxFileRepository"]


# Initiate logger.
log = logging.getLogger("dodecahedron")


class AbstractXlsxFileRepository(AbstractFileSystemRepository):
    """Represents an abstract `.xlsx` file repository.

    Args:
        wrapper: Xlsx file wrapper.
        *args (optional): Positional arguments.
        mapper (optional): Mapper. Default ``None``.
        **kwargs (optional): Keyword arguments.

    Attributes:
        columns: Columns names.

    """

    def __init__(
        self,
        wrapper: Any,
        /,
        *args: Any,
        mapper: Optional["ClassMapper[Any]"] = None,
        **kwargs: Any,
    ) -> None:
        if not isinstance(wrapper, XlsxFileWrapper):
            expected = "expected type 'XlsxFileWrapper'"
            actual = f"got {type(wrapper)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        if mapper and not isinstance(mapper, ClassMapper):  # type: ignore
            expected = "expected type 'ClassMapper'"
            actual = f"got {type(mapper)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(wrapper, *args, **kwargs)
        self._mapper = mapper

    @property
    def columns(self) -> Optional[Sequence[Any]]:
        """Column names."""
        results = getattr(self._wrapper, "fieldnames", None)
        return results

    @property
    def mapper(self) -> Optional["ClassMapper[Any]"]:
        """Mapper."""
        return self._mapper

    def _read_records(self) -> List[Dict[Hashable, Any]]:
        """Read records from file.

        Returns:
            Records.

        """
        with self._wrapper.open() as file:
            results = getattr(file, "read_records")()

        return results

    def _write_records(self, records: List[Dict[Hashable, Any]], /) -> None:
        """Write records to file.

        Args:
            records: Records.

        """
        with self._wrapper.open("wb") as file:
            getattr(file, "write_header")()
            results = getattr(file, "write_records")(records)

        return results



# -*- coding: utf-8 -*-

# Standard Library Imports
import abc
import functools
from types import FunctionType
from types import MethodType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar
import warnings

__all__ = ["TrackerMeta"]


# Constants
ADD_METHOD = "add"
GET_METHOD = "get"
LIST_METHOD = "list"
REMOVE_METHOD = "remove"
SEEN_ATTR = "__seen__"

# Custom types
T = TypeVar("T")


class TrackerMeta(abc.ABCMeta):
    """Metaclass for tracking child objects."""

    def __new__(
        meta,  # type: ignore
        name: str,
        bases: Tuple[type, ...],
        namespace: Dict[str, Any],
        **kwargs: Any,
    ) -> type:
        wrapped_attrs = meta.wrap_attributes(namespace)
        return super().__new__(meta, name, bases, wrapped_attrs, **kwargs)

    def __call__(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        instance = super().__call__(*args, **kwargs)
        setattr(instance, SEEN_ATTR, set())
        return instance

    @classmethod
    def wrap_attributes(cls, attrs: Dict[str, Any]) -> Dict[str, Any]:
        """Wrap attributes.

         Args:
            attrs: Attributes to wrap.

        Returns:
            Wrapped attributes.

        """
        results = {
            key: (
                cls.wrap_method(key, value)
                if isinstance(value, (FunctionType, MethodType))
                else value
            )
            for key, value in attrs.items()
        }
        return results

    @classmethod
    def wrap_method(cls, name: str, method: Any) -> Callable[..., Any]:
        """Wrap method.

        Args:
            method: Method to wrap.

        Returns:
            Wrapped method.

        """
        if not isinstance(method, (FunctionType, MethodType)):
            message = f"expected method, got type {type(method)} instead"
            raise TypeError(message)

        if name == ADD_METHOD:
            return track_first_positional_argument(method)

        if name == GET_METHOD:
            return track_single_return_value(method)

        if name == LIST_METHOD:
            return track_multiple_return_values(method)

        if name == REMOVE_METHOD:
            return track_first_positional_argument(method)

        return method


def track_first_positional_argument(
    method: Callable[..., Any], /
) -> Callable[..., Any]:
    """Track first positional argument passed to method.

    Args:
        method: Method to track.

    Returns:
        Wrapped function.

    Raises:
        TypeError: when argument is not a method.

    """
    if not isinstance(method, (FunctionType, MethodType)):
        message = f"expected method, got type {type(method)} instead"
        raise TypeError(message)

    @functools.wraps(method)
    def wrapper(self: Any, obj: object, /, *args: Any, **kwargs: Any) -> Any:
        """Wrapper applied to decorated method.

        Args:
            obj: Object to track.
            *args: Positional arguments to pass to wrapped method.
            **kwargs: Keyword arguments to pass to wrapped method.

        Returns:
            Result of called method.

        """
        result = method(self, obj, *args, **kwargs)

        # We only add objects to those seen after the method executes
        # successfully. We don't want to track objects that raise exceptions.
        add_seen_object(self, obj)
        return result

    functools.update_wrapper(wrapper, method)
    return wrapper


def track_multiple_return_values(
    method: Callable[..., Any], /
) -> Callable[..., Any]:
    """Track all values returned from method.

    Args:
        method: Method to track.

    Returns:
        Wrapped function.

    Raises:
        TypeError: when argument is not a method.

    """
    if not isinstance(method, (FunctionType, MethodType)):
        message = f"expected method, got type {type(method)} instead"
        raise TypeError(message)

    @functools.wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Sequence[Any]:
        """Wrapper applied to decorated method.

        Args:
            *args: Positional arguments to pass to wrapped method.
            **kwargs: Keyword arguments to pass to wrapped method.

        Returns:
            Results of called method.

        """
        results = method(self, *args, **kwargs)
        if results is not None:
            update_seen_objects(self, results)

        return results

    functools.update_wrapper(wrapper, method)
    return wrapper


def track_single_return_value(
    method: Callable[..., Any], /
) -> Callable[..., Any]:
    """Track each value returned from method.

    Args:
        method: Method to track.

    Returns:
        Wrapped function.

    Raises:
        TypeError: when argument is not a method.

    """
    if not isinstance(method, (FunctionType, MethodType)):
        message = f"expected method, got type {type(method)} instead"
        raise TypeError(message)

    @functools.wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        """Wrapper applied to decorated method.

        Args:
            *args: Positional arguments to pass to wrapped method.
            **kwargs: Keyword arguments to pass to wrapped method.

        Returns:
            Result of called method.

        """
        result = method(self, *args, **kwargs)
        if result is not None:
            add_seen_object(self, result)

        return result

    functools.update_wrapper(wrapper, method)
    return wrapper


# ----------------------------------------------------------------------------
# Helper Functions
# ----------------------------------------------------------------------------
def add_seen_object(parent: Any, child: Any) -> None:
    """Add value to seen objects.

    Args:
        parent: Parent object.
        child: Child object seen.

    """
    try:
        set_default_attr(parent, SEEN_ATTR, set())
        seen: Set[Any] = getattr(parent, SEEN_ATTR)
        seen.add(child)

    except TypeError:
        message = "Tracking decorator does not support unhashable objects"
        warnings.warn(message)


def update_seen_objects(parent: Any, children: Sequence[Any]) -> None:
    """Add results to seen objects.

    Args:
        parent: Parent object.
        children: Child objects seen.

    """
    try:
        set_default_attr(parent, SEEN_ATTR, set())
        seen: Set[Any] = getattr(parent, SEEN_ATTR)
        seen.update(children)

    except TypeError:
        message = "Tracking decorator does not support unhashable objects"
        warnings.warn(message)


def set_default_attr(obj: object, attr: str, value: Any) -> None:
    """Set default value of attribute on object.

    Args:
        obj: Object on which to set default attribute.
        attr: Name of attribute to set.
        value: Value to which to set attribute.

    """
    if not hasattr(obj, attr):
        setattr(obj, attr, value)



# -*- coding: utf-8 -*-
"""Command-Line Interface

The command-line interface is a wrapper around the argparse library which receives
arguments that the user provides and calls the appropriate service function.

"""

# pylint: disable=no-self-use

# Standard Library Imports
import abc
from argparse import ArgumentParser
from argparse import Namespace
import inspect
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union


# Initialize logger.
log = logging.getLogger("dodecahedron")

# Custom types
CommandLineArguments = Union[Namespace, Sequence[str]]

# Constants
PROCESS = "process"


class CommandLineInterface(abc.ABC):
    """Class implements a command-line interface (CLI).

    Args:
        parser: Parser.
        processes: Processes.

    Attributes:
        parser: Argument parser.

    """

    def __init__(
        self,
        parser: ArgumentParser,
        /,
        processes: Dict[str, Callable[..., None]],
    ) -> None:
        log.debug("Starting command-line interface...")
        self._parser = parser
        self._processes = processes
        self._parser.add_argument(
            PROCESS,
            type=str,
            choices=list(self._processes.keys()),
            help=f"{PROCESS} to run",
        )
        # TODO: Add arguments to parser from function signatures
        self._parser.set_defaults(func=self.execute)

    @property
    def parser(self) -> ArgumentParser:
        """Argument parser."""
        return self._parser

    @property
    def processes(self) -> Dict[str, Callable[..., None]]:
        """Processes."""
        return self._processes  # pragma: no cover

    def __call__(self, args: Namespace) -> None:
        self.execute(args)  # pragma: no cover

    def execute(
        self,
        args: Optional[CommandLineArguments] = None,
        /,
    ) -> None:
        """Execute function based on the provided argument(s).

        Args:
            args: Arguments passed to command line.

        """
        namespace = self._parse_arguments(args)
        process: str = getattr(namespace, PROCESS)
        self._execute_process(process.lower(), namespace)

    def _parse_arguments(
        self,
        args: Optional[CommandLineArguments],
        /,
    ) -> Namespace:
        """Parse command-line arguments.

        Args:
            args: Arguments passed to command line.

        """
        result = (
            self._parser.parse_args(args)
            if not isinstance(args, Namespace)
            else args
        )
        return result

    def _execute_process(
        self,
        __process: str,
        __namespace: Namespace,
        /,
    ) -> None:
        """Execute process.

        Args:
            __process: Name of process.
            __namespace: Namespace.

        """
        log.debug("Executing '%s' process...", __process)
        process = self._processes[__process]
        signature = inspect.signature(process)
        args = select_positional_arguments(__namespace, signature)
        kwargs = select_keyword_arguments(__namespace, signature)
        process(*args, **kwargs)


# ----------------------------------------------------------------------------
# Selectors
# ----------------------------------------------------------------------------
def select_positional_arguments(
    __namespace: Namespace,
    __signature: inspect.Signature,
    /,
) -> Tuple[Any, ...]:
    """Select positional arguments.

    Args:
        __namespace: Namespace.
        __signature: Signature.

    Returns:
        Positional Arguments.

    """
    results = get_positional_only_arguments(__namespace, __signature)
    return results


def select_keyword_arguments(
    __namespace: Namespace,
    __signature: inspect.Signature,
    /,
) -> Dict[str, Any]:
    """Select keyword arguments.

    Args:
        __namespace: Namespace.
        __signature: Signature.

    Returns:
        Keyword Arguments.

    """
    results: Dict[str, Any] = {
        **get_positional_or_keyword_arguments(__namespace, __signature),
        **get_keyword_only_arguments(__namespace, __signature),
    }
    return results


def get_positional_only_arguments(
    __namespace: Namespace,
    __signature: inspect.Signature,
    /,
) -> Tuple[Any, ...]:
    """Get positional-only arguments.

    Args:
        __namespace: Namespace.
        __signature: Signature.

    Returns:
        Positional-Only Arguments.

    """
    params: List[str] = [
        name
        for name, param in __signature.parameters.items()
        if param.kind.name == "POSITIONAL_ONLY"
    ]
    results = tuple(getattr(__namespace, param) for param in params)
    return results


def get_positional_or_keyword_arguments(
    __namespace: Namespace,
    __signature: inspect.Signature,
    /,
) -> Dict[str, Any]:
    """Get positional or keyword arguments.

    Args:
        __namespace: Namespace.
        __signature: Signature.

    Returns:
        Positional-or-Keyword Arguments.

    """
    params: List[str] = [
        name
        for name, param in __signature.parameters.items()
        if param.kind.name == "POSITIONAL_OR_KEYWORD"
    ]
    results = {param: getattr(__namespace, param) for param in params}
    return results


def get_keyword_only_arguments(
    __namespace: Namespace,
    __signature: inspect.Signature,
    /,
) -> Dict[str, Any]:
    """Get keyword-only arguments.

    Args:
        __namespace: Namespace.
        __signature: Signature.

    Returns:
        Keyword-Only Arguments.

    """
    params: List[str] = [
        name
        for name, param in __signature.parameters.items()
        if param.kind.name == "KEYWORD_ONLY"
    ]
    results = {param: getattr(__namespace, param, None) for param in params}
    return results



# -*- coding: utf-8 -*-
"""Repositories.

Todo:
    * Add 'sqlite3' repository.

"""

# Standard Library Imports
import importlib

# Local Imports
from .abstract_repository import *
from .csv_repository import *
from .eventful_repository import *
from .txt_repository import *

try:
    importlib.import_module("sqlalchemy")  # pragma: no cover
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:  # pragma: no cover
    from .sqlalchemy_repository import *

try:
    importlib.import_module("openpyxl")  # pragma: no cover
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:  # pragma: no cover
    from .xlsx_repository import *
